"""Essentially contains the definition of

1. The forward model (an polynomial regression of a grid of simulations of the Meudon PDR code)
2. the negative logpdf of the likelihood, prior and posterior, their gradient and the diagonal of their hessian matrix
"""
